"""Core functionality for querying the Apple location service."""

import logging
import struct
from dataclasses import asdict, dataclass
from typing import List

import requests

from gsloc.protobuf.location_pb2 import Request, Response  # pylint: disable=no-name-in-module

logger = logging.getLogger(__name__)


class NoResultsError(Exception):
    """Raised when no results are returned from the Apple location service."""

    def __init__(self, mac: str):
        super().__init__(f"No results found for MAC address: {mac}")


@dataclass
class WifiInfo:
    """Information about a WiFi network."""

    mac: str
    channel: int
    latitude: float
    longitude: float
    accuracy: int
    altitude: int
    altitude_accuracy: int

    def validate(self) -> None:
        """Validates the WifiInfo object."""
        assert 0 <= self.channel <= 255, "Channel must be between 0 and 255."
        assert -90 <= self.latitude <= 90, "Latitude must be between -90 and 90."
        assert -180 <= self.longitude <= 180, "Longitude must be between -180 and 180."
        assert self.accuracy >= 0, "Accuracy must be greater than or equal to 0."
        assert self.altitude >= 0, "Altitude must be greater than or equal to 0."
        assert self.altitude_accuracy >= 0, "Altitude accuracy must be greater than or equal to 0."

    @classmethod
    def from_response_wifi(cls, response_wifi: Response.ResponseWifi) -> "WifiInfo":
        """Creates a WifiInfo object from a ResponseWifi object."""
        try:
            return cls(
                mac=response_wifi.mac,
                channel=response_wifi.channel,
                latitude=response_wifi.location.latitude / 1e8,
                longitude=response_wifi.location.longitude / 1e8,
                accuracy=response_wifi.location.accuracy,
                altitude=response_wifi.location.altitude,
                altitude_accuracy=response_wifi.location.altitudeAccuracy,
            )
        except AssertionError as e:
            # Fields are invalid if Apple doesn't know this wifi access point
            raise NoResultsError(response_wifi.mac) from e

    def to_feature(self) -> dict:
        """Converts the WifiInfo object to a GeoJSON feature."""
        return {
            "type": "Feature",
            "geometry": {
                "type": "Point",
                "coordinates": [self.longitude, self.latitude],
            },
            "properties": asdict(self),
        }


@dataclass
class PayloadHeader:
    """Header for the request's payload."""

    nul_sqh = b"\x00\x01"
    nul_nul = b"\x00\x00"
    llength = b"\x00\x05"
    locale = b"\x65\x6e\x5f\x55\x53"
    ilength = b"\x00\x13"
    identifier = b"\x63\x6f\x6d\x2e\x61\x70\x70\x6c\x65\x2e\x6c\x6f\x63\x61\x74\x69\x6f\x6e\x64"
    vlength = b"\x00\x0c"
    version = b"\x38\x2e\x34\x2e\x31\x2e\x31\x32\x48\x33\x32\x31"

    def build(self) -> bytes:
        """Builds the header for the request's payload."""
        return b"".join(
            (
                self.nul_sqh,
                self.llength,
                self.locale,
                self.ilength,
                self.identifier,
                self.vlength,
                self.version,
                self.nul_nul,
                self.nul_sqh,
                self.nul_nul,
            )
        )


class Payload:
    """Payload for the request."""

    def __init__(
        self, header: PayloadHeader, mac_addresses: List[str], noise: int = 0, signal: int = 100
    ):
        self.header = header
        self.request = Request()
        self.request.noise = noise
        self.request.signal = signal
        for mac in mac_addresses:
            self.request.wifis.add(mac=mac)

    def build(self) -> bytes:
        """Builds the payload for the request."""
        message = self.request.SerializeToString()
        size = struct.pack(">h", len(message))
        return self.header.build() + size + message


def query(macs: List[str]) -> List[WifiInfo]:
    """Queries the Apple location service with the given payload."""
    logger.info("Querying Apple location service with MAC addresses: %s", macs)
    payload = Payload(PayloadHeader(), macs)
    response = requests.post(
        url="https://gs-loc.apple.com/clls/wloc",
        data=payload.build(),
        headers={
            "User-Agent": "locationd/221 (com.apple.locationd/1.0)",
            "Content-Type": "application/octet-stream",
        },
        timeout=60,
    )
    raw_data = response.content
    if not raw_data:
        logger.error("Empty response received from Apple location service.")
        return []
    data_buffer = raw_data[raw_data.find(b"\x00\x00\x00\x01\x00\x00") + 8 :]
    response = Response()
    response.ParseFromString(data_buffer)

    logger.info("Received response from Apple location service.")

    return [WifiInfo.from_response_wifi(wifi) for wifi in response.wifis]



"""Command-line interface for querying the Apple location service."""

import argparse
import json
import logging
import sys
from pathlib import Path

import rich
from rich.logging import RichHandler

from gsloc.core import NoResultsError, WifiInfo, query

logger = logging.getLogger(__name__)


def configure_logging(verbose: bool = False):
    """Configures logging for the application."""
    level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        level=level,
        handlers=[RichHandler()],
    )


def parse_arguments() -> argparse.Namespace:
    """Parses command-line arguments."""
    parser = argparse.ArgumentParser(
        description="Query the Apple location service for WiFi network locations."
    )
    parser.add_argument(
        "mac_addresses",
        nargs="*",
        help="MAC addresses to query. If not provided, reads from stdin.",
    )
    parser.add_argument(
        "-i",
        "--input",
        type=Path,
        help="Path to a file containing MAC addresses to query, one per line.",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        help="Path to a file to write the results to in GeoJson format.",
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Enable debug logging.")
    return parser.parse_args()


def read_mac_addresses_from_file(file_path: Path) -> list[str]:
    """Reads MAC addresses from a file, one per line."""
    with file_path.open() as f:
        return [line.strip() for line in f]


def write_results_to_file(file_path: Path, results: list[WifiInfo]) -> None:
    """Writes the results to a file in GeoJSON format."""
    features = [result.to_feature() for result in results]

    geojson = {
        "type": "FeatureCollection",
        "features": features,
    }

    with file_path.open("w") as f:
        json.dump(geojson, f, indent=2)


def main() -> None:
    """Main entry point for the application."""
    args = parse_arguments()
    configure_logging(args.verbose)

    if args.input:
        mac_addresses = read_mac_addresses_from_file(args.input)
    elif args.mac_addresses:
        mac_addresses = args.mac_addresses
    else:
        mac_addresses = [line.strip() for line in sys.stdin]

    results: list[WifiInfo] = []
    for mac in mac_addresses:
        try:
            wifi_list = query([mac])
            if wifi_list:
                results.extend(wifi_list)
        except NoResultsError as e:
            logger.warning(str(e))

    if args.output:
        write_results_to_file(args.output, results)
    else:
        rich.print(results)



# type: ignore
# pylint: disable=all
from typing import ClassVar as _ClassVar
from typing import Iterable as _Iterable
from typing import Mapping as _Mapping
from typing import Optional as _Optional
from typing import Union as _Union

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf.internal import containers as _containers

DESCRIPTOR: _descriptor.FileDescriptor

class Response(_message.Message):
    __slots__ = ("wifis",)

    class ResponseWifi(_message.Message):
        __slots__ = ("mac", "location", "channel")

        class WifiLocation(_message.Message):
            __slots__ = (
                "latitude",
                "longitude",
                "accuracy",
                "zeroField4",
                "altitude",
                "altitudeAccuracy",
                "unknown11",
                "unknown12",
            )
            LATITUDE_FIELD_NUMBER: _ClassVar[int]
            LONGITUDE_FIELD_NUMBER: _ClassVar[int]
            ACCURACY_FIELD_NUMBER: _ClassVar[int]
            ZEROFIELD4_FIELD_NUMBER: _ClassVar[int]
            ALTITUDE_FIELD_NUMBER: _ClassVar[int]
            ALTITUDEACCURACY_FIELD_NUMBER: _ClassVar[int]
            UNKNOWN11_FIELD_NUMBER: _ClassVar[int]
            UNKNOWN12_FIELD_NUMBER: _ClassVar[int]
            latitude: int
            longitude: int
            accuracy: int
            zeroField4: int
            altitude: int
            altitudeAccuracy: int
            unknown11: int
            unknown12: int
            def __init__(
                self,
                latitude: _Optional[int] = ...,
                longitude: _Optional[int] = ...,
                accuracy: _Optional[int] = ...,
                zeroField4: _Optional[int] = ...,
                altitude: _Optional[int] = ...,
                altitudeAccuracy: _Optional[int] = ...,
                unknown11: _Optional[int] = ...,
                unknown12: _Optional[int] = ...,
            ) -> None: ...

        MAC_FIELD_NUMBER: _ClassVar[int]
        LOCATION_FIELD_NUMBER: _ClassVar[int]
        CHANNEL_FIELD_NUMBER: _ClassVar[int]
        mac: str
        location: Response.ResponseWifi.WifiLocation
        channel: int
        def __init__(
            self,
            mac: _Optional[str] = ...,
            location: _Optional[_Union[Response.ResponseWifi.WifiLocation, _Mapping]] = ...,
            channel: _Optional[int] = ...,
        ) -> None: ...

    WIFIS_FIELD_NUMBER: _ClassVar[int]
    wifis: _containers.RepeatedCompositeFieldContainer[Response.ResponseWifi]
    def __init__(
        self, wifis: _Optional[_Iterable[_Union[Response.ResponseWifi, _Mapping]]] = ...
    ) -> None: ...

class Request(_message.Message):
    __slots__ = ("wifis", "noise", "signal", "source")

    class RequestWifi(_message.Message):
        __slots__ = ("mac",)
        MAC_FIELD_NUMBER: _ClassVar[int]
        mac: str
        def __init__(self, mac: _Optional[str] = ...) -> None: ...

    WIFIS_FIELD_NUMBER: _ClassVar[int]
    NOISE_FIELD_NUMBER: _ClassVar[int]
    SIGNAL_FIELD_NUMBER: _ClassVar[int]
    SOURCE_FIELD_NUMBER: _ClassVar[int]
    wifis: _containers.RepeatedCompositeFieldContainer[Request.RequestWifi]
    noise: int
    signal: int
    source: str
    def __init__(
        self,
        wifis: _Optional[_Iterable[_Union[Request.RequestWifi, _Mapping]]] = ...,
        noise: _Optional[int] = ...,
        signal: _Optional[int] = ...,
        source: _Optional[str] = ...,
    ) -> None: ...



# -*- coding: utf-8 -*-
# type: ignore
# pylint: disable=all
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: location.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0elocation.proto"\xc5\x02\n\x08Response\x12%\n\x05wifis\x18\x02 \x03(\x0b\x32\x16.Response.ResponseWifi\x1a\x91\x02\n\x0cResponseWifi\x12\x0b\n\x03mac\x18\x01 \x01(\t\x12\x35\n\x08location\x18\x02 \x01(\x0b\x32#.Response.ResponseWifi.WifiLocation\x12\x0f\n\x07\x63hannel\x18\x15 \x01(\x05\x1a\xab\x01\n\x0cWifiLocation\x12\x10\n\x08latitude\x18\x01 \x01(\x03\x12\x11\n\tlongitude\x18\x02 \x01(\x03\x12\x10\n\x08\x61\x63\x63uracy\x18\x03 \x01(\x05\x12\x12\n\nzeroField4\x18\x04 \x01(\x05\x12\x10\n\x08\x61ltitude\x18\x05 \x01(\x05\x12\x18\n\x10\x61ltitudeAccuracy\x18\x06 \x01(\x05\x12\x11\n\tunknown11\x18\x0b \x01(\x05\x12\x11\n\tunknown12\x18\x0c \x01(\x05"\x81\x01\n\x07Request\x12#\n\x05wifis\x18\x02 \x03(\x0b\x32\x14.Request.RequestWifi\x12\x10\n\x05noise\x18\x03 \x01(\x05:\x01\x30\x12\x13\n\x06signal\x18\x04 \x01(\x05:\x03\x31\x30\x30\x12\x0e\n\x06source\x18\x05 \x01(\t\x1a\x1a\n\x0bRequestWifi\x12\x0b\n\x03mac\x18\x01 \x01(\t'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "location_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_RESPONSE"]._serialized_start = 19
    _globals["_RESPONSE"]._serialized_end = 344
    _globals["_RESPONSE_RESPONSEWIFI"]._serialized_start = 71
    _globals["_RESPONSE_RESPONSEWIFI"]._serialized_end = 344
    _globals["_RESPONSE_RESPONSEWIFI_WIFILOCATION"]._serialized_start = 173
    _globals["_RESPONSE_RESPONSEWIFI_WIFILOCATION"]._serialized_end = 344
    _globals["_REQUEST"]._serialized_start = 347
    _globals["_REQUEST"]._serialized_end = 476
    _globals["_REQUEST_REQUESTWIFI"]._serialized_start = 450
    _globals["_REQUEST_REQUESTWIFI"]._serialized_end = 476
# @@protoc_insertion_point(module_scope)
